import os
import traceback
from typing import Callable, Any, Optional

from .client import ManageClient


def log_exception(
    definition: dict,
    description: str,
    trb: Optional[str] = None,
    function_name: Optional[str] = '',
    is_auth: bool = False,
) -> bool:
    """Log error helper

    Args:
        definition (dict): dict definiton
        description (str): exception str
        trb (Optional[str], optional): full traceback. Defaults to None.
        function_name (Optional[str], optional): function. Defaults to ''.

    Returns:
        bool: status
    """
    init_kwargs = {
        "system_id": os.environ.get("MANAGE_SYSTEM_ID", ""),
        "system_url": os.environ.get("MANAGE_SYSTEM_URL", ""),
        "token": os.environ.get("MANAGE_SYSTEM_TOKEN", ""),
        "debug": os.environ.get("MANAGE_SYSTEM_DEBUG", '0'),
    }
    client = ManageClient(**init_kwargs)
    if not trb:
        trb = str(traceback.format_exc()).replace(
            'func_result = function', function_name
        )
    return client.log_exception(definition, description, trb=trb, is_auth=is_auth)


def run(function: Callable[..., Any], **kwargs) -> Any:
    """
    function must be a top level function, not class method, and not decorated function
    """
    try:
        from zappa.asynchronous import run as zappa_run
    except ImportError:
        raise ImportError('u need to install zappa==0.50.0')
    if "kwargs" in kwargs:
        kwargs = kwargs["kwargs"]
    if not os.environ.get("AWS_LAMBDA_FUNCTION_NAME"):
        return function(**kwargs)
    zappa_run(function, kwargs=kwargs)
    return True



import os
import traceback
import datetime
from typing import Callable, Union, Any

from .utils import get_list_values, get_values, _check_lambda_args
from .client import ManageClient

__all__ = ('event', 'list_value_event')


class __Executor(object):
    def __init__(
        self,
        function: Callable,
        keys=None,
        list_value=None,
        iter_key=None,
        description=None,
        *args,
        **kwargs
    ):
        self.function = function
        self.keys = keys
        self.list_value = list_value
        self.iter_key = iter_key
        self.args = args
        self.kwargs = kwargs
        self.description = description

    def execution(self, definition: dict, timestamp: str, *args, **kwargs) -> Any:
        init_kwargs = {
            "system_id": os.environ.get("MANAGE_SYSTEM_ID", ""),
            "system_url": os.environ.get("MANAGE_SYSTEM_URL", ""),
            "token": os.environ.get("MANAGE_SYSTEM_TOKEN", ""),
            "debug": os.environ.get("MANAGE_SYSTEM_DEBUG", '0'),
        }
        client = ManageClient(**init_kwargs)
        client.push_event(
            definition=definition,
            start_time=timestamp,
            status=False,
            description=self.description,
        )
        # print(pushed)
        try:
            args = _check_lambda_args(args)
            if args or kwargs:
                func_result = self.function(*args, **kwargs)
            else:
                func_result = self.function()
        except Exception as exc_description:
            t = traceback.format_exc()
            t = t.replace("func_result = function", self.function.__name__)
            return client.log_exception(
                definition=definition, description=str(exc_description), trb=t
            )
        client.push_event(
            definition=definition,
            start_time=timestamp,
            status=True,
            description=self.description,
        )
        return func_result

    def push(self, list_: bool = False, *args, **kwargs):
        definition = {}
        var_names = self.function.__code__.co_varnames
        try:
            if self.keys is not None and not list_:
                values = get_values(self.keys, var_names, *args, **kwargs)
                definition = values["definition"]
                if not definition:
                    definition = {self.function.__name__: self.function.__name__}
            elif list_:
                list_values = get_list_values(
                    self.keys,
                    self.list_value,
                    self.iter_key,
                    var_names,
                    *args,
                    **kwargs
                )
                definition = list_values["definition"]
            else:
                definition = {self.function.__name__: self.function.__name__}

        except IndexError:
            pass
        timestamp = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        return self.execution(definition, timestamp, *args, **kwargs)


def event(keys=None, description=None):
    def decorator(function, *args, **kwargs):
        def push_function_results(*args, **kwargs):
            executor = __Executor(function=function, keys=keys, description=description)
            return executor.push(False, *args, **kwargs)

        return push_function_results

    return decorator


def list_value_event(
    keys: Union[str, list], list_value: str, iter_key: str, description=None
) -> Any:
    def decorator(function, *args, **kwargs):
        def push_function_results(*args, **kwargs):
            executor = __Executor(
                function=function,
                keys=keys,
                list_value=list_value,
                iter_key=iter_key,
                description=description,
            )
            return executor.push(True, *args, **kwargs)

        return push_function_results

    return decorator



import os


class ManageClientConfig(object):
    def __init__(
        self,
        manage_system_id: str,
        manage_system_url: str,
        manage_system_token: str,
        manage_system_debug: bool = False,
    ):
        self.manage_system_id = manage_system_id
        self.manage_system_url = manage_system_url
        self.manage_system_token = manage_system_token
        self.manage_system_debug = '1' if manage_system_debug else '0'
        self.init_environ_settings()

    def init_environ_settings(self) -> None:
        os.environ["MANAGE_SYSTEM_ID"] = self.manage_system_id
        os.environ["MANAGE_SYSTEM_URL"] = self.manage_system_url
        os.environ["MANAGE_SYSTEM_TOKEN"] = self.manage_system_token
        os.environ["MANAGE_SYSTEM_DEBUG"] = self.manage_system_debug



from typing import Optional, Union
from json import loads, JSONDecodeError
from smart_crypt import Crypt


def decrypt_request_data(data: dict, salt: str, token: str) -> dict:
    c = Crypt(salt=salt)
    try:
        decrypted_str = c.decrypt(data['data'], token)
        return loads(decrypted_str)
    except (ValueError, JSONDecodeError):
        raise ValueError('cant decrypt data, invalid salt or token')


def _create_definition(keys: Union[str, list], data: dict) -> dict:
    """return name/title for event/error

    Args:
        keys (Union[str, list]): keys for chacke in data
        data (dict): function data

    Returns:
        dict: name/title
    """
    if isinstance(keys, str):
        return {keys: data[keys]}
    elif isinstance(keys, (list, tuple)):
        return {k: data[k] for k in keys}
    else:
        return {}


# CREATE EVENT data
def get_list_values(
    keys: Union[str, list],
    list_value: str,
    iter_key: str,
    var_names: list,
    *args,
    **kwargs,
) -> dict:
    """return name and description

    Args:
        keys (Union[str, list]):  keys for chacke in data
        list_value (str): task list 
        iter_key (str): iter key
        var_names (list): func names

    Returns:
        dict: definiton
    """
    list_ = []
    index = 0
    if list_value in kwargs:
        list_ = kwargs[list_value]
    elif list_value in var_names:
        list_ = args[var_names.index(list_value)]
    if iter_key in kwargs:
        index = kwargs[iter_key]
    elif iter_key in var_names:
        index = args[var_names.index(iter_key)]
    data = list_[index]
    definition = _create_definition(keys, data)
    return {
        "definition": definition,
    }


# for creating default name
def get_values(keys: Union[str, list], var_names: list, *args, **kwargs,) -> dict:
    """ger values

    Args:
        keys (Union[str, list]):  keys for chacke in data
        var_names (list): func var names

    Returns:
        dict: definition
    """
    definition = {}
    if isinstance(keys, str):
        if keys in kwargs:
            definition[keys] = kwargs[keys]
        elif keys in var_names:
            definition[keys] = args[var_names.index(keys)]
        else:
            definition[keys] = ''
    else:
        for key in keys:
            if key in kwargs:
                value = kwargs[key]
            elif key in var_names:
                value = args[var_names.index(key)]
            else:
                raise ValueError('invalid keys')
            definition[key] = value
    return {"definition": definition}


def create_event_definition(
    keys: Union[str, list],
    values: Union[dict, list, None] = None,
    key: Optional[int] = None,
) -> str:
    """helper for create event name, error na,e

    Args:
        keys (Union[str, list]):  keys for chacke in data
        values (Union[dict, list, None], optional): data. Defaults to None.
        key (Optional[int], optional): [description]. Defaults to None.

    Returns:
        str: {"definition": <definiton>}
    """
    if isinstance(values, list) and isinstance(key, int):
        data = values[key]
    else:
        data = values
    try:
        return get_values(keys, [], **data)["definition"]
    except IndexError:
        pass


def _check_lambda_args(args):
    for arg in args:
        if arg.__class__.__name__ == "LambdaContext":
            return tuple()
    return args



from setuptools import setup, find_packages

setup(
    name="smart-manage-app-client",
    version='0.0.7b',
    packages=find_packages(),
    install_requires=["requests>=2.18.2", "pytz", "smart-manage-crypt"],
)



import warnings
import time
import datetime
import logging
from typing import Optional
from requests import post, ConnectionError

logger = logging.getLogger('manage_app_client')


def _request(url: str, data: dict, headers: dict) -> bool:
    """Requests helper

    Args:
        url (str): manage system url
        data (dict): event data or exceptio data
        headers (dict): auth headers

    Returns:
        bool: True if success else False
    """
    try:
        r = post(url, json=data, headers=headers)
        if r.status_code > 204:
            warnings.warn(
                f'Invalid response status: {r.status_code}, check Manage app settings'
            )
            return False
        return True
    except ConnectionError as e:
        return False
    except Exception as e:
        return False


class ManageClient(object):
    def __init__(
        self, token: str, system_id: str, system_url: str, debug: bool = False
    ) -> None:
        """Init manage client

        Args:
            token (str): secure token
            system_id (str): system id uuid4
            system_url (str): manage app backend url
            debug (bool, optional): manage deug. Defaults to False.
        """
        self.token = token
        self.system_id = system_id
        self.system_url = self._create_system_url(system_url)
        self.debug = True if int(debug) else False

    def _create_system_url(self, url: str) -> str:
        if not url.endswith("/api/event/"):
            if url.endswith("/"):
                url = f"{url}"
            else:
                url = f"{url}/"
        return url

    def push_event(
        self,
        definition: dict,
        start_time: str,
        status: bool = False,
        description: Optional[str] = None,
    ) -> bool:
        """Push event data

        Args:
            definition (dict): definition of event
            status (bool, optional): finisth status. Defaults to False.

        Returns:
            bool: status
        """

        headers = {"execute-token": self.token}
        name = ', '.join(f'{key}: {value}' for key, value in definition.items())
        if not name:
            return False
        data = {
            "title": name,
            "definition": definition,
            "description": description,
            "is_finish": status,
        }
        if status:
            time.sleep(0.1)
            data["start_time"] = start_time
            data["finish_time"] = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        else:
            data["start_time"] = start_time
            data["finish_time"] = None
        return _request(
            f'{self.system_url}api/events/?system_id={self.system_id}',
            data=data,
            headers=headers,
        )

    def push_exception(
        self,
        definition: str,
        description: str,
        start_time: str,
        is_auth: bool = False,
        trb: Optional[str] = None,
    ) -> bool:
        """Push error

        Args:
            definition (str): like {"integration_id": 13, "selection": 2310}
            description (str): ZeroDivizionError
            is_auth (bool, optional): [description]. Defaults to False.
            trb (Optional[str], optional): full traceback. Defaults to None.

        Returns:
            bool: status
        """
        if self.debug:
            logger.debug(trb)
            print(trb)
        else:
            logger.info(trb)
        exception_url = f'{self.system_url}api/errors/?system_id={self.system_id}'
        name = ', '.join(f'{key}: {value}' for key, value in definition.items())
        headers = {"execute-token": self.token}
        data = {
            "name": name,
            "title": name,
            "description": description,
            "definition": definition,
            "traceback": trb if trb else description,
            "date_time": start_time,
            "is_auth": is_auth,
        }
        _request(exception_url, data=data, headers=headers)
        return True

    def log_exception(
        self,
        definition: dict,
        description: str,
        is_auth: bool = False,
        trb: Optional[str] = None,
    ) -> bool:
        """Puth to errors enpoint

        Args:
            definition (str): event definition
            trb (Optional[str], optional): traceback. Defaults to None.

        Returns:
            bool: status
        """
        dt = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        self.push_event(
            definition=definition, start_time=dt, status=True, description=description,
        )
        if self.debug:
            return False
        return self.push_exception(definition, description, dt, is_auth, trb)
